@@ -16,7 +16,7 @@
   GNU General Public License for more details.
 
 */
-/* $Id: client_notify.c,v 1.37 2002/03/24 14:18:42 priikone Exp $ */
+/* $Id: client_notify.c,v 1.53 2002/05/16 11:33:09 priikone Exp $ */
 /* This file includes the Notify packet handling. Notify packets are
    important packets sent by the server. They tell different things to the
    client such as nick changes, mode changes etc. */
@@ -25,12 +25,25 @@
 #include "silcclient.h"
 #include "client_internal.h"
 
+/* Context used for resolving client, channel and server info. */
 typedef struct {
-  SilcPacketContext *packet;
+  void *packet;
   void *context;
   SilcSocketConnection sock;
 } *SilcClientNotifyResolve;
 
+SILC_TASK_CALLBACK(silc_client_notify_check_client)
+{ 
+  SilcClientNotifyResolve res = (SilcClientNotifyResolve)context;
+  SilcClient client = res->context;
+  SilcClientConnection conn = res->sock->user_data;
+  SilcClientID *client_id = res->packet;
+  silc_client_get_client_by_id_resolve(client, conn, client_id, NULL, NULL);
+  silc_free(client_id);
+  silc_socket_free(res->sock);
+  silc_free(res);
+}
+
 /* Called when notify is received and some async operation (such as command)
    is required before processing the notify message. This calls again the
    silc_client_notify_by_server and reprocesses the original notify packet. */
@@ -43,11 +56,8 @@ static void silc_client_notify_by_server_pending(void *context, void *context2)
 
   SILC_LOG_DEBUG(("Start"));
 
-  if (reply) {
-    SilcCommandStatus status = silc_command_get_status(reply->payload);
-    if (status != SILC_STATUS_OK)
-      goto out;
-  }
+  if (reply && !silc_command_get_status(reply->payload, NULL, NULL))
+    goto out;
 
   silc_client_notify_by_server(res->context, res->sock, res->packet);
 
@@ -72,13 +82,13 @@ static void silc_client_notify_by_server_resolve(SilcClient client,
   res->context = client;
   res->sock = silc_socket_dup(conn->sock);
 
-  /* For client resolving use WHOIS, and oterhwise use IDENTIFY */
+  /* For client resolving use WHOIS, and otherwise use IDENTIFY */
   if (id_type == SILC_ID_CLIENT) {
     silc_client_command_register(client, SILC_COMMAND_WHOIS, NULL, NULL,
 				 silc_client_command_reply_whois_i, 0,
 				 ++conn->cmd_ident);
     silc_client_command_send(client, conn, SILC_COMMAND_WHOIS, conn->cmd_ident,
-			     1, 3, idp->data, idp->len);
+			     1, 4, idp->data, idp->len);
     silc_client_command_pending(conn, SILC_COMMAND_WHOIS, conn->cmd_ident,
 				silc_client_notify_by_server_pending, res);
   } else {
@@ -214,9 +224,10 @@ void silc_client_notify_by_server(SilcClient client,
 	client_entry->status &= ~SILC_CLIENT_STATUS_RESOLVING;
 	goto out;
       }
-      client_entry->status |= SILC_CLIENT_STATUS_RESOLVING;
       silc_client_notify_by_server_resolve(client, conn, packet, 
 					   SILC_ID_CLIENT, client_id);
+      client_entry->status |= SILC_CLIENT_STATUS_RESOLVING;
+      client_entry->resolve_cmd_ident = conn->cmd_ident;
       goto out;
     } else {
       if (client_entry != conn->local_entry)
@@ -292,6 +303,21 @@ void silc_client_notify_by_server(SilcClient client,
       silc_free(chu);
     }
 
+    /* Some client implementations actually quit network by first doing
+       LEAVE and then immediately SIGNOFF.  We'll check for this by doing 
+       check for the client after 5 - 34 seconds.  If it is not valid after
+       that we'll remove the client from cache. */
+    if (!silc_hash_table_count(client_entry->channels)) {
+      SilcClientNotifyResolve res = silc_calloc(1, sizeof(*res));
+      res->context = client;
+      res->sock = silc_socket_dup(conn->sock);
+      res->packet = silc_id_dup(client_id, SILC_ID_CLIENT);
+      silc_schedule_task_add(client->schedule, conn->sock->sock,
+			     silc_client_notify_check_client, res,
+			     (5 + (silc_rng_get_rn16(client->rng) % 29)),
+			     0, SILC_TASK_TIMEOUT, SILC_TASK_PRI_NORMAL);
+    }
+
     /* Notify application. The channel entry is sent last as this notify
        is for channel but application don't know it from the arguments
        sent by server. */
@@ -433,7 +459,7 @@ void silc_client_notify_by_server(SilcClient client,
       goto out;
 
     /* Ignore my ID */
-    if (SILC_ID_CLIENT_COMPARE(client_id, conn->local_id))
+    if (conn->local_id && SILC_ID_CLIENT_COMPARE(client_id, conn->local_id))
       break;
 
     /* Find old Client entry */
@@ -453,41 +479,96 @@ void silc_client_notify_by_server(SilcClient client,
     if (!client_id)
       goto out;
 
-    /* Find Client entry and if not found resolve it */
-    client_entry2 = silc_client_get_client_by_id(client, conn, client_id);
-    if (!client_entry2) {
-      /* Resolve the entry information */
-      silc_client_notify_by_server_resolve(client, conn, packet, 
-					   SILC_ID_CLIENT, client_id);
+    /* From protocol version 1.1 we get the new nickname in notify as well,
+       so we don't have to resolve it.  Do it the hard way if server doesn't
+       send it to us. */
+    tmp = silc_argument_get_arg_type(args, 3, NULL);
+    if (tmp) {
+      /* Protocol version 1.1 */
+      char *tmp_nick = NULL;
 
-      /* Add the new entry even though we resolved it. This is because we
-	 want to replace the old entry with the new entry here right now. */
-      client_entry2 = 
-	silc_client_add_client(client, conn, NULL, NULL, NULL, 
-			       silc_id_dup(client_id, SILC_ID_CLIENT), 
-			       client_entry->mode);
+      /* Check whether nickname changed at all.  It is possible that nick
+	 change notify is received but nickname didn't changed, only the
+	 ID changes. */
+      if (client->internal->params->nickname_parse)
+	client->internal->params->nickname_parse(client_entry->nickname,
+						 &tmp_nick);
+      else
+	tmp_nick = strdup(tmp);
 
-      /* Replace old ID entry with new one on all channels. */
-      silc_client_replace_from_channels(client, conn, client_entry,
-					client_entry2);
+      if (tmp_nick && !strcmp(tmp, tmp_nick)) {
+	/* Nickname didn't change. Update only the ID */
+	silc_idcache_del_by_context(conn->client_cache, client_entry);
+	silc_free(client_entry->id);
+	client_entry->id = silc_id_dup(client_id, SILC_ID_CLIENT);
+	silc_idcache_add(conn->client_cache, strdup(tmp),
+			 client_entry->id, client_entry, 0, NULL);
+
+	/* Notify application */
+	client->internal->ops->notify(client, conn, type, 
+				      client_entry, client_entry);
+	break;
+      }
+      silc_free(tmp_nick);
+
+      /* Create new client entry, and save all old information with the
+	 new nickname and client ID */
+      client_entry2 = silc_client_add_client(client, conn, NULL, NULL, 
+					     client_entry->realname,
+					     silc_id_dup(client_id, 
+							 SILC_ID_CLIENT), 0);
+      if (!client_entry2)
+	goto out;
+
+      if (client_entry->server)
+	client_entry2->server = strdup(client_entry->server);
+      if (client_entry->username)
+	client_entry2->username = strdup(client_entry->username);
+      if (client_entry->hostname)
+	client_entry2->hostname = strdup(client_entry->hostname);
+      silc_client_update_client(client, conn, client_entry2, tmp, NULL, NULL,
+				client_entry->mode);
     } else {
+      /* Protocol version 1.0 */
+
+      /* Find client entry and if not found resolve it */
+      client_entry2 = silc_client_get_client_by_id(client, conn, client_id);
+      if (!client_entry2) {
+	/* Resolve the entry information */
+	silc_client_notify_by_server_resolve(client, conn, packet, 
+					     SILC_ID_CLIENT, client_id);
+
+	/* Add the new entry even though we resolved it. This is because we
+	   want to replace the old entry with the new entry here right now. */
+	client_entry2 = 
+	  silc_client_add_client(client, conn, NULL, NULL, NULL, 
+				 silc_id_dup(client_id, SILC_ID_CLIENT), 
+				 client_entry->mode);
+
+	/* Replace old ID entry with new one on all channels. */
+	silc_client_replace_from_channels(client, conn, client_entry,
+					  client_entry2);
+	break;
+      }
+
       if (client_entry2 != conn->local_entry)
 	silc_client_nickname_format(client, conn, client_entry2);
-
-      /* Remove the old from cache */
-      silc_idcache_del_by_context(conn->client_cache, client_entry);
-
-      /* Replace old ID entry with new one on all channels. */
-      silc_client_replace_from_channels(client, conn, client_entry,
-					client_entry2);
-
-      /* Notify application */
-      client->internal->ops->notify(client, conn, type, 
-				    client_entry, client_entry2);
-
-      /* Free data */
-      silc_client_del_client_entry(client, conn, client_entry);
     }
+
+    /* Remove the old from cache */
+    silc_idcache_del_by_context(conn->client_cache, client_entry);
+    
+    /* Replace old ID entry with new one on all channels. */
+    silc_client_replace_from_channels(client, conn, client_entry,
+				      client_entry2);
+
+    /* Notify application */
+    client->internal->ops->notify(client, conn, type, 
+				  client_entry, client_entry2);
+    
+    /* Free old client entry */
+    silc_client_del_client_entry(client, conn, client_entry);
+
     break;
 
   case SILC_NOTIFY_TYPE_CMODE_CHANGE:
@@ -769,9 +850,10 @@ void silc_client_notify_by_server(SilcClient client,
     if (!channel)
       break;
 
-    /* Get the kicker */
+    /* From protocol version 1.1 we get the kicker's client ID as well */
     tmp = silc_argument_get_arg_type(args, 3, &tmp_len);
     if (tmp) {
+      silc_free(client_id);
       client_id = silc_id_payload_parse_id(tmp, tmp_len, NULL);
       if (!client_id)
 	goto out;
@@ -797,50 +879,116 @@ void silc_client_notify_by_server(SilcClient client,
     client->internal->ops->notify(client, conn, type, client_entry, tmp, 
 				  client_entry2, channel);
 
-    /* If I was kicked from channel, remove the channel */
+    /* Remove kicked client from channel */
     if (client_entry == conn->local_entry) {
+      /* If I was kicked from channel, remove the channel */
       if (conn->current_channel == channel)
 	conn->current_channel = NULL;
-      silc_idcache_del_by_id(conn->channel_cache, channel->id);
-      silc_free(channel->channel_name);
-      silc_free(channel->id);
-      silc_free(channel->key);
-      silc_cipher_free(channel->channel_key);
-      silc_free(channel);
+      silc_client_del_channel(client, conn, channel);
+    } else {
+      chu = silc_client_on_channel(channel, client_entry);
+      if (chu) {
+	silc_hash_table_del(client_entry->channels, channel);
+	silc_hash_table_del(channel->user_list, client_entry);
+	silc_free(chu);
+      }
+
+      if (!silc_hash_table_count(client_entry->channels)) {
+	SilcClientNotifyResolve res = silc_calloc(1, sizeof(*res));
+	res->context = client;
+	res->sock = silc_socket_dup(conn->sock);
+	res->packet = silc_id_dup(client_entry->id, SILC_ID_CLIENT);
+	silc_schedule_task_add(client->schedule, conn->sock->sock,
+			       silc_client_notify_check_client, res,
+			       (5 + (silc_rng_get_rn16(client->rng) % 529)),
+			       0, SILC_TASK_TIMEOUT, SILC_TASK_PRI_NORMAL);
+      }
     }
     break;
 
   case SILC_NOTIFY_TYPE_KILLED:
-    /*
-     * A client (maybe me) was killed from the network.
-     */
+    {
+      /*
+       * A client (maybe me) was killed from the network.
+       */
+      char *comment;
+      SilcUInt32 comment_len;
 
-    SILC_LOG_DEBUG(("Notify: KILLED"));
+      SILC_LOG_DEBUG(("Notify: KILLED"));
 
-    /* Get Client ID */
-    tmp = silc_argument_get_arg_type(args, 1, &tmp_len);
-    if (!tmp)
-      goto out;
+      /* Get Client ID */
+      tmp = silc_argument_get_arg_type(args, 1, &tmp_len);
+      if (!tmp)
+	goto out;
 
-    client_id = silc_id_payload_parse_id(tmp, tmp_len, NULL);
-    if (!client_id)
-      goto out;
+      client_id = silc_id_payload_parse_id(tmp, tmp_len, NULL);
+      if (!client_id)
+	goto out;
 
-    /* Find Client entry */
-    client_entry = silc_client_get_client_by_id(client, conn, client_id);
-    if (!client_entry)
-      goto out;
+      /* Find Client entry */
+      client_entry = silc_client_get_client_by_id(client, conn, client_id);
+      if (!client_entry)
+	goto out;
 
-    /* Get comment */
-    tmp = silc_argument_get_arg_type(args, 2, &tmp_len);
+      /* Get comment */
+      comment = silc_argument_get_arg_type(args, 2, &comment_len);
 
-    /* Notify application. */
-    client->internal->ops->notify(client, conn, type, client_entry, tmp);
+      /* From protocol version 1.1 we get killer's client ID as well */
+      tmp = silc_argument_get_arg_type(args, 3, &tmp_len);
+      if (tmp) {
+	silc_free(client_id);
+	id = silc_id_payload_parse_id(tmp, tmp_len, &id_type);
+	if (!id)
+	  goto out;
 
-    if (client_entry != conn->local_entry)
-      /* Remove the client from all channels and free it */
-      silc_client_del_client(client, conn, client_entry);
+	/* Find Client entry */
+	if (id_type == SILC_ID_CLIENT) {
+	  /* Find Client entry */
+	  client_id = id;
+	  client_entry2 = silc_client_get_client_by_id(client, conn, 
+						       client_id);
+	  if (!client_entry) {
+	    silc_client_notify_by_server_resolve(client, conn, packet, 
+						 SILC_ID_CLIENT, client_id);
+	    goto out;
+	  }
+	} else if (id_type == SILC_ID_SERVER) {
+	  /* Find Server entry */
+	  server_id = id;
+	  server = silc_client_get_server_by_id(client, conn, server_id);
+	  if (!server) {
+	    silc_client_notify_by_server_resolve(client, conn, packet, 
+						 SILC_ID_SERVER, server_id);
+	    goto out;
+	  }
+      
+	  /* Save the pointer to the client_entry pointer */
+	  client_entry2 = (SilcClientEntry)server;
+	} else {
+	  /* Find Channel entry */
+	  channel_id = id;
+	  channel = silc_client_get_channel_by_id(client, conn, channel_id);
+	  if (!channel) {
+	    silc_client_notify_by_server_resolve(client, conn, packet, 
+						 SILC_ID_CHANNEL, channel_id);
+	    goto out;
+	  }
+	  
+	  /* Save the pointer to the client_entry pointer */
+	  client_entry2 = (SilcClientEntry)channel;
+	  silc_free(channel_id);
+	  channel_id = NULL;
+	}
+      }
 
+      /* Notify application. */
+      client->internal->ops->notify(client, conn, type, client_entry, 
+				    comment, id_type, client_entry2);
+
+      if (client_entry != conn->local_entry)
+	/* Remove the client from all channels and free it */
+	silc_client_del_client(client, conn, client_entry);
+    }
     break;
     
   case SILC_NOTIFY_TYPE_SERVER_SIGNOFF:
@@ -896,6 +1044,113 @@ void silc_client_notify_by_server(SilcClient client,
     }
     break;
 
+  case SILC_NOTIFY_TYPE_ERROR:
+    {
+      /*
+       * Some has occurred and server is notifying us about it.
+       */
+      SilcStatus error;
+
+      tmp = silc_argument_get_arg_type(args, 1, &tmp_len);
+      if (!tmp && tmp_len != 1)
+	goto out;
+      error = (SilcStatus)tmp[0];
+
+      SILC_LOG_DEBUG(("Notify: ERROR (%d)", error));
+
+      if (error == SILC_STATUS_ERR_NO_SUCH_CLIENT_ID) {
+	tmp = silc_argument_get_arg_type(args, 2, &tmp_len);
+	if (tmp) {
+	  client_id = silc_id_payload_parse_id(tmp, tmp_len, NULL);
+	  if (!client_id)
+	    goto out;
+	  client_entry = silc_client_get_client_by_id(client, conn,
+						      client_id);
+	  if (client_entry)
+	    silc_client_del_client(client, conn, client_entry);
+	}
+      }
+
+      /* Notify application. */
+      client->internal->ops->notify(client, conn, type, error);
+    }
+    break;
+
+  case SILC_NOTIFY_TYPE_WATCH:
+    {
+      /*
+       * Received notify about some client we are watching
+       */
+      SilcNotifyType notify = 0;
+
+      SILC_LOG_DEBUG(("Notify: WATCH"));
+
+      /* Get sender Client ID */
+      tmp = silc_argument_get_arg_type(args, 1, &tmp_len);
+      if (!tmp)
+	goto out;
+      client_id = silc_id_payload_parse_id(tmp, tmp_len, NULL);
+      if (!client_id)
+	goto out;
+
+      /* Find Client entry and if not found query it */
+      client_entry = silc_client_get_client_by_id(client, conn, client_id);
+      if (!client_entry) {
+	silc_client_notify_by_server_resolve(client, conn, packet, 
+					     SILC_ID_CLIENT, client_id);
+	goto out;
+      }
+
+      /* Get user mode */
+      tmp = silc_argument_get_arg_type(args, 3, &tmp_len);
+      if (!tmp || tmp_len != 4)
+	goto out;
+      SILC_GET32_MSB(mode, tmp);
+
+      /* Get notify type */
+      tmp = silc_argument_get_arg_type(args, 4, &tmp_len);
+      if (tmp && tmp_len != 2)
+	goto out;
+      if (tmp)
+	SILC_GET16_MSB(notify, tmp);
+
+      /* Get nickname */
+      tmp = silc_argument_get_arg_type(args, 2, NULL);
+      if (tmp) {
+	char *tmp_nick = NULL;
+
+	if (client->internal->params->nickname_parse)
+	  client->internal->params->nickname_parse(client_entry->nickname,
+						   &tmp_nick);
+	else
+	  tmp_nick = strdup(tmp);
+
+	/* If same nick, the client was new to us and has become "present"
+	   to network.  Send NULL as nick to application. */
+	if (!strcmp(tmp, tmp_nick))
+	  tmp = NULL;
+
+	silc_free(tmp_nick);
+      }
+
+      /* Notify application. */
+      client->internal->ops->notify(client, conn, type, client_entry,
+				    tmp, mode, notify);
+
+      client_entry->mode = mode;
+
+      /* If nickname was changed, remove the client entry unless the
+	 client is on some channel */
+      if (tmp && notify == SILC_NOTIFY_TYPE_NICK_CHANGE &&
+	  !silc_hash_table_count(client_entry->channels))
+	silc_client_del_client(client, conn, client_entry);
+      else if (notify == SILC_NOTIFY_TYPE_SIGNOFF ||
+	       notify == SILC_NOTIFY_TYPE_SERVER_SIGNOFF ||
+	       notify == SILC_NOTIFY_TYPE_KILLED)
+	silc_client_del_client(client, conn, client_entry);
+    }
+    break;
+
   default:
     break;
   }
