@@ -1,6 +1,6 @@
 /*
 
-  silcutil.c 
+  silcutil.c
 
   Author: Pekka Riikonen <priikone@silcnet.org>
 
@@ -20,7 +20,7 @@
  * These are general utility functions that doesn't belong to any specific
  * group of routines.
  */
-/* $Id: silcutil.c,v 1.41 2002/03/24 14:18:42 priikone Exp $ */
+/* $Id: silcutil.c,v 1.48 2002/04/21 17:59:12 priikone Exp $ */
 
 #include "silcincludes.h"
 
@@ -32,27 +32,27 @@ int silc_gets(char *dest, int destlen, const char *src, int srclen, int begin)
 {
   static int start = 0;
   int i;
-  
+
   memset(dest, 0, destlen);
-  
+
   if (begin != start)
     start = 0;
-  
+
   i = 0;
   for ( ; start <= srclen; i++, start++) {
     if (i > destlen)
       return -1;
-    
+
     dest[i] = src[start];
-    
-    if (dest[i] == EOF) 
+
+    if (dest[i] == EOF)
       return EOF;
-    
-    if (dest[i] == '\n') 
+
+    if (dest[i] == '\n')
       break;
   }
   start++;
-  
+
   return start;
 }
 
@@ -60,7 +60,7 @@ int silc_gets(char *dest, int destlen, const char *src, int srclen, int begin)
    were found. This is used to check for bad lines when reading data from
    for example a configuration file. */
 
-int silc_check_line(char *buf) 
+int silc_check_line(char *buf)
 {
   /* Illegal characters in line */
   if (strchr(buf, '#')) return -1;
@@ -70,11 +70,11 @@ int silc_check_line(char *buf)
   if (strchr(buf, '\a')) return -1;
   if (strchr(buf, '\b')) return -1;
   if (strchr(buf, '\f')) return -1;
-  
+
   /* Empty line */
   if (buf[0] == '\n')
     return -1;
-  
+
   return 0;
 }
 
@@ -92,171 +92,37 @@ char *silc_get_time()
   return return_time;
 }
 
-/* Converts string to capital characters */
+/* Converts string to capital characters. */
 
-char *silc_to_upper(char *string)
+bool silc_to_upper(const char *string, char *dest, SilcUInt32 dest_size)
 {
   int i;
-  char *ret = silc_calloc(strlen(string) + 1, sizeof(char));
+
+  if (strlen(string) > dest_size)
+    return FALSE;
 
   for (i = 0; i < strlen(string); i++)
-    ret[i] = toupper(string[i]);
+    dest[i] = toupper(string[i]);
 
-  return ret;
+  return TRUE;
 }
 
-static unsigned char pem_enc[64] =
-"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
+/* Converts string to lower letter characters. */
 
-/* Encodes data into PEM encoding. Returns NULL terminated PEM encoded
-   data string. Note: This is originally public domain code and is 
-   still PD. */
-
-char *silc_encode_pem(unsigned char *data, SilcUInt32 len)
+bool silc_to_lower(const char *string, char *dest, SilcUInt32 dest_size)
 {
-  int i, j;
-  SilcUInt32 bits, c, char_count;
-  char *pem;
+  int i;
 
-  char_count = 0;
-  bits = 0;
-  j = 0;
+  if (strlen(string) > dest_size)
+    return FALSE;
 
-  pem = silc_calloc(((len * 8 + 5) / 6) + 5, sizeof(*pem));
+  for (i = 0; i < strlen(string); i++)
+    dest[i] = tolower(string[i]);
 
-  for (i = 0; i < len; i++) {
-    c = data[i];
-    bits += c;
-    char_count++;
-
-    if (char_count == 3) {
-      pem[j++] = pem_enc[bits  >> 18];
-      pem[j++] = pem_enc[(bits >> 12) & 0x3f];
-      pem[j++] = pem_enc[(bits >> 6)  & 0x3f];
-      pem[j++] = pem_enc[bits & 0x3f];
-      bits = 0;
-      char_count = 0;
-    } else {
-      bits <<= 8;
-    }
-  }
-
-  if (char_count != 0) {
-    bits <<= 16 - (8 * char_count);
-    pem[j++] = pem_enc[bits >> 18];
-    pem[j++] = pem_enc[(bits >> 12) & 0x3f];
-
-    if (char_count == 1) {
-      pem[j++] = '=';
-      pem[j] = '=';
-    } else {
-      pem[j++] = pem_enc[(bits >> 6) & 0x3f];
-      pem[j] = '=';
-    }
-  }
-
-  return pem;
+  return TRUE;
 }
 
-/* Same as above but puts newline ('\n') every 72 characters. */
-
-char *silc_encode_pem_file(unsigned char *data, SilcUInt32 data_len)
-{
-  int i, j;
-  SilcUInt32 len, cols;
-  char *pem, *pem2;
-
-  pem = silc_encode_pem(data, data_len);
-  len = strlen(pem);
-
-  pem2 = silc_calloc(len + (len / 72) + 1, sizeof(*pem2));
-
-  for (i = 0, j = 0, cols = 1; i < len; i++, cols++) {
-    if (cols == 72) {
-      pem2[i] = '\n';
-      cols = 0;
-      len++;
-      continue;
-    }
-
-    pem2[i] = pem[j++];
-  }
-
-  silc_free(pem);
-  return pem2;
-}
-
-/* Decodes PEM into data. Returns the decoded data. Note: This is
-   originally public domain code and is still PD. */
-
-unsigned char *silc_decode_pem(unsigned char *pem, SilcUInt32 pem_len,
-			       SilcUInt32 *ret_len)
-{
-  int i, j;
-  SilcUInt32 len, c, char_count, bits;
-  unsigned char *data;
-  static char ialpha[256], decoder[256];
-
-  for (i = 64 - 1; i >= 0; i--) {
-    ialpha[pem_enc[i]] = 1;
-    decoder[pem_enc[i]] = i;
-  }
-
-  char_count = 0;
-  bits = 0;
-  j = 0;
-
-  if (!pem_len)
-    len = strlen(pem);
-  else
-    len = pem_len;
-
-  data = silc_calloc(((len * 6) / 8), sizeof(*data));
-
-  for (i = 0; i < len; i++) {
-    c = pem[i];
-
-    if (c == '=')
-      break;
-
-    if (c > 127 || !ialpha[c])
-      continue;
-
-    bits += decoder[c];
-    char_count++;
-
-    if (char_count == 4) {
-      data[j++] = bits >> 16;
-      data[j++] = (bits >> 8) & 0xff;
-      data[j++] = bits & 0xff;
-      bits = 0;
-      char_count = 0;
-    } else {
-      bits <<= 6;
-    }
-  }
-
-  switch(char_count) {
-  case 1:
-    silc_free(data);
-    return NULL;
-    break;
-  case 2:
-    data[j++] = bits >> 10;
-    break;
-  case 3:
-    data[j++] = bits >> 16;
-    data[j++] = (bits >> 8) & 0xff;
-    break;
-  }
-
-  if (ret_len)
-    *ret_len = j;
-
-  return data;
-}
-
-/* Parse userfqdn string which is in user@fqdn format */
+/* Parse userfqdn string which is in user@fqdn format. */
 
 bool silc_parse_userfqdn(const char *string, char **left, char **right)
 {
@@ -273,12 +139,12 @@ bool silc_parse_userfqdn(const char *string, char **left, char **right)
 
   if (strchr(string, '@')) {
     tlen = strcspn(string, "@");
-    
+
     if (left) {
       *left = silc_calloc(tlen + 1, sizeof(char));
       memcpy(*left, string, tlen);
     }
-    
+
     if (right) {
       *right = silc_calloc((strlen(string) - tlen) + 1, sizeof(char));
       memcpy(*right, string + tlen + 1, strlen(string) - tlen - 1);
@@ -296,7 +162,7 @@ bool silc_parse_userfqdn(const char *string, char **left, char **right)
    arguments in the line. Note that the command name is counted as one
    argument and is saved. */
 
-void silc_parse_command_line(unsigned char *buffer, 
+void silc_parse_command_line(unsigned char *buffer,
 			     unsigned char ***parsed,
 			     SilcUInt32 **parsed_lens,
 			     SilcUInt32 **parsed_types,
@@ -313,7 +179,10 @@ void silc_parse_command_line(unsigned char *buffer,
 
   /* Get the command first */
   len = strcspn(cp, " ");
-  tmp = silc_to_upper((char *)cp);
+  tmp = silc_calloc(strlen(cp) + 1, sizeof(*tmp));
+  if (!tmp)
+    return;
+  silc_to_upper(cp, tmp, strlen(cp));
   (*parsed)[0] = silc_calloc(len + 1, sizeof(char));
   memcpy((*parsed)[0], tmp, len);
   silc_free(tmp);
@@ -335,9 +204,9 @@ void silc_parse_command_line(unsigned char *buffer,
 	len--;
       if (!len)
 	break;
-      
+
       *parsed = silc_realloc(*parsed, sizeof(**parsed) * (argc + 1));
-      *parsed_lens = silc_realloc(*parsed_lens, 
+      *parsed_lens = silc_realloc(*parsed_lens,
 				  sizeof(**parsed_lens) * (argc + 1));
       (*parsed)[argc] = silc_calloc(len + 1, sizeof(char));
       memcpy((*parsed)[argc], cp, len);
@@ -382,15 +251,15 @@ char *silc_format(char *fmt, ...)
 /* Renders ID to suitable to print for example to log file. */
 
 static char rid[256];
-#define _PUT_STRING(d, s)				\
-do {							\
-  int sp = (sizeof(d) - 1) - strlen(d);			\
-  if (sp < strlen(s)) {					\
-    if (sp)						\
-      strncat(d, s, (sizeof(d) - 1) - strlen(d));	\
-  } else {						\
-    strncat(d, s, strlen(s));				\
-  }							\
+#define _PUT_STRING(__d__, __s__)					\
+do {									\
+  int __sp = sizeof(__d__) - 1 - strlen(__d__);				\
+  if (__sp < strlen(__s__)) {						\
+    if (__sp)								\
+      strncat(__d__, __s__, (sizeof(__d__) - 1) - strlen(__d__));	\
+  } else {								\
+    strncat(__d__, __s__, strlen(__s__));				\
+  }									\
 } while(0)
 
 char *silc_id_render(void *id, SilcUInt16 type)
@@ -456,7 +325,7 @@ char *silc_id_render(void *id, SilcUInt16 type)
       snprintf(tmp, sizeof(tmp) - 1, ",%02x,", client_id->rnd);
       _PUT_STRING(rid, tmp);
       memset(tmp, 0, sizeof(tmp));
-      snprintf(tmp, sizeof(tmp) - 1, "[%02x %02x %02x %02x...]", 
+      snprintf(tmp, sizeof(tmp) - 1, "[%02x %02x %02x %02x...]",
 	       client_id->hash[0], client_id->hash[1],
 	       client_id->hash[2], client_id->hash[3]);
       _PUT_STRING(rid, tmp);
@@ -496,8 +365,9 @@ char *silc_id_render(void *id, SilcUInt16 type)
 
   return rid;
 }
+#undef _PUT_STRING
 
-/* Compares two strings. Strings may include wildcards * and ?.
+/* Compares two strings. Strings may include wildcards '*' and '?'.
    Returns TRUE if strings match. */
 
 int silc_string_compare(char *string1, char *string2)
@@ -520,15 +390,15 @@ int silc_string_compare(char *string1, char *string2)
   if (slen2 < slen1)
     if (!strchr(string1, '*'))
       return FALSE;
-  
+
   /* Take copies of the original strings as we will change them */
   tmpstr1 = silc_calloc(slen1 + 1, sizeof(char));
   memcpy(tmpstr1, string1, slen1);
   tmpstr2 = silc_calloc(slen2 + 1, sizeof(char));
   memcpy(tmpstr2, string2, slen2);
-  
+
   for (i = 0; i < slen1; i++) {
-    
+
     /* * wildcard. Only one * wildcard is possible. */
     if (tmpstr1[i] == '*')
       if (!strncmp(tmpstr1, tmpstr2, i)) {
@@ -536,7 +406,7 @@ int silc_string_compare(char *string1, char *string2)
 	strncpy(tmpstr2, tmpstr1, i);
 	break;
       }
-    
+
     /* ? wildcard */
     if (tmpstr1[i] == '?') {
       if (!strncmp(tmpstr1, tmpstr2, i)) {
@@ -544,17 +414,17 @@ int silc_string_compare(char *string1, char *string2)
 	  if (tmpstr1[i + 1] != '?' &&
 	      tmpstr1[i + 1] != tmpstr2[i + 1])
 	    continue;
-	
+
 	if (!(slen1 < slen2))
 	  tmpstr2[i] = '?';
       }
     }
   }
-  
+
   /* if using *, remove it */
   if (strchr(tmpstr1, '*'))
     *strchr(tmpstr1, '*') = 0;
-  
+
   if (!strcmp(tmpstr1, tmpstr2)) {
     memset(tmpstr1, 0, slen1);
     memset(tmpstr2, 0, slen2);
@@ -562,7 +432,7 @@ int silc_string_compare(char *string1, char *string2)
     silc_free(tmpstr2);
     return TRUE;
   }
-  
+
   memset(tmpstr1, 0, slen1);
   memset(tmpstr2, 0, slen2);
   silc_free(tmpstr1);
@@ -570,7 +440,7 @@ int silc_string_compare(char *string1, char *string2)
   return FALSE;
 }
 
-/* Basic has function to hash strings. May be used with the SilcHashTable. 
+/* Basic has function to hash strings. May be used with the SilcHashTable.
    Note that this lowers the characters of the string (with tolower()) so
    this is used usually with nicknames, channel and server names to provide
    case insensitive keys. */
@@ -579,7 +449,7 @@ SilcUInt32 silc_hash_string(void *key, void *user_context)
 {
   char *s = (char *)key;
   SilcUInt32 h = 0, g;
-  
+
   while (*s != '\0') {
     h = (h << 4) + tolower(*s);
     if ((g = h & 0xf0000000)) {
@@ -588,7 +458,7 @@ SilcUInt32 silc_hash_string(void *key, void *user_context)
     }
     s++;
   }
-  
+
   return h;
 }
 
@@ -618,43 +488,33 @@ SilcUInt32 silc_hash_id(void *key, void *user_context)
   case SILC_ID_CLIENT:
     {
       SilcClientID *id = (SilcClientID *)key;
-      SilcUInt32 g;
-  
-      /* The client ID is hashed by hashing the hash of the ID 
+
+      /* The client ID is hashed by hashing the hash of the ID
 	 (which is a truncated MD5 hash of the nickname) so that we
 	 can access the entry from the cache with both Client ID but
 	 with just a hash from the ID as well. */
-
-      for (i = 0; i < sizeof(id->hash); i++) {
-	h = (h << 4) + id->hash[i];
-	if ((g = h & 0xf0000000)) {
-	  h = h ^ (g >> 24);
-	  h = h ^ g;
-	}
-      }
-
-      return h;
+      return silc_hash_client_id_hash(id->hash, NULL);
     }
     break;
   case SILC_ID_SERVER:
     {
       SilcServerID *id = (SilcServerID *)key;
-      
+
       h = id->port * id->rnd;
       for (i = 0; i < id->ip.data_len; i++)
 	h ^= id->ip.data[i];
-      
+
       return h;
     }
     break;
   case SILC_ID_CHANNEL:
     {
       SilcChannelID *id = (SilcChannelID *)key;
-      
+
       h = id->port * id->rnd;
       for (i = 0; i < id->ip.data_len; i++)
 	h ^= id->ip.data[i];
-      
+
       return h;
     }
     break;
@@ -665,6 +525,25 @@ SilcUInt32 silc_hash_id(void *key, void *user_context)
   return h;
 }
 
+/* Hash Client ID's hash. */
+
+SilcUInt32 silc_hash_client_id_hash(void *key, void *user_context)
+{
+  int i;
+  unsigned char *hash = key;
+  SilcUInt32 h = 0, g;
+
+  for (i = 0; i < CLIENTID_HASH_LEN; i++) {
+    h = (h << 4) + hash[i];
+    if ((g = h & 0xf0000000)) {
+      h = h ^ (g >> 24);
+      h = h ^ g;
+    }
+  }
+
+  return h;
+}
+
 /* Hash binary data. The `user_context' is the data length. */
 
 SilcUInt32 silc_hash_data(void *key, void *user_context)
@@ -690,21 +569,22 @@ SilcUInt32 silc_hash_public_key(void *key, void *user_context)
 	  silc_hash_data(pk->pk, (void *)pk->pk_len));
 }
 
-/* Compares two strings. May be used as SilcHashTable comparison function. */
+/* Compares two strings. It may be used as SilcHashTable comparison
+   function. */
 
 bool silc_hash_string_compare(void *key1, void *key2, void *user_context)
 {
   return !strcasecmp((char *)key1, (char *)key2);
 }
 
-/* Compares two ID's. May be used as SilcHashTable comparison function. 
+/* Compares two ID's. May be used as SilcHashTable comparison function.
    The Client ID's compares only the hash of the Client ID not any other
    part of the Client ID. Other ID's are fully compared. */
 
 bool silc_hash_id_compare(void *key1, void *key2, void *user_context)
 {
   SilcIdType id_type = (SilcIdType)(SilcUInt32)user_context;
-  return (id_type == SILC_ID_CLIENT ? 
+  return (id_type == SILC_ID_CLIENT ?
 	  SILC_ID_COMPARE_HASH((SilcClientID *)key1, (SilcClientID *)key2) :
 	  SILC_ID_COMPARE_TYPE(key1, key2, id_type));
 }
@@ -724,8 +604,8 @@ bool silc_hash_data_compare(void *key1, void *key2, void *user_context)
   return !memcmp(key1, key2, len);
 }
 
-/* Compares two SILC Public keys. May be used as SilcHashTable comparison
-   function. */
+/* Compares two SILC Public keys. It may be used as SilcHashTable
+   comparison function. */
 
 bool silc_hash_public_key_compare(void *key1, void *key2, void *user_context)
 {
@@ -767,6 +647,12 @@ char *silc_client_chmode(SilcUInt32 mode, const char *cipher, const char *hmac)
   if (mode & SILC_CHANNEL_MODE_FOUNDER_AUTH)
     strncat(string, "f", 1);
 
+  if (mode & SILC_CHANNEL_MODE_SILENCE_USERS)
+    strncat(string, "m", 1);
+
+  if (mode & SILC_CHANNEL_MODE_SILENCE_OPERS)
+    strncat(string, "M", 1);
+
   if (mode & SILC_CHANNEL_MODE_CIPHER)
     strncat(string, cipher, strlen(cipher));
 
@@ -795,6 +681,15 @@ char *silc_client_chumode(SilcUInt32 mode)
   if (mode & SILC_CHANNEL_UMODE_CHANOP)
     strncat(string, "o", 1);
 
+  if (mode & SILC_CHANNEL_UMODE_BLOCK_MESSAGES)
+    strncat(string, "b", 1);
+
+  if (mode & SILC_CHANNEL_UMODE_BLOCK_MESSAGES_USERS)
+    strncat(string, "u", 1);
+
+  if (mode & SILC_CHANNEL_UMODE_BLOCK_MESSAGES_ROBOTS)
+    strncat(string, "r", 1);
+
   return strdup(string);
 }
 
@@ -830,7 +725,7 @@ char *silc_fingerprint(const unsigned char *data, SilcUInt32 data_len)
   for (i = 0; i < data_len; i++) {
     snprintf(cp, sizeof(fingerprint), "%02X", data[i]);
     cp += 2;
-    
+
     if ((i + 1) % 2 == 0)
       snprintf(cp++, sizeof(fingerprint), " ");
 
@@ -842,7 +737,7 @@ char *silc_fingerprint(const unsigned char *data, SilcUInt32 data_len)
     cp[-2] = 0;
   if ((i + 1) % 10 == 0)
     cp[-1] = 0;
-  
+
   return strdup(fingerprint);
 }
 
@@ -865,7 +760,7 @@ bool silc_string_is_ascii(const unsigned char *data, SilcUInt32 data_len)
 bool silc_parse_version_string(const char *version,
 			       SilcUInt32 *protocol_version,
 			       char **protocol_version_string,
-			       SilcUInt32 *software_version, 
+			       SilcUInt32 *software_version,
 			       char **software_version_string,
 			       char **vendor_version)
 {
@@ -899,7 +794,7 @@ bool silc_parse_version_string(const char *version,
 
   /* Take software version */
 
-  maj = 0; 
+  maj = 0;
   min = 0;
   cp = strchr(cp, '-');
   if (!cp)
@@ -1040,3 +935,99 @@ char *silc_get_input(const char *prompt, bool echo_off)
   return NULL;
 #endif /* SILC_UNIX */
 }
+
+/* Return mode list */
+
+bool silc_get_mode_list(SilcBuffer mode_list, SilcUInt32 mode_list_count,
+			SilcUInt32 **list)
+{
+  int i;
+
+  if (mode_list->len / 4 != mode_list_count)
+    return FALSE;
+
+  *list = silc_calloc(mode_list_count, sizeof(**list));
+
+  for (i = 0; i < mode_list_count; i++) {
+    SILC_GET32_MSB((*list)[i], mode_list->data);
+    silc_buffer_pull(mode_list, 4);
+  }
+
+  silc_buffer_push(mode_list, mode_list->data - mode_list->head);
+
+  return TRUE;
+}
+
+/* Status message structure. Messages are defined below. */
+typedef struct {
+  SilcStatus status;
+  char *message;
+} SilcStatusMessage;
+
+#define STAT(x) SILC_STATUS_ERR_##x
+const SilcStatusMessage silc_status_messages[] = {
+
+  { STAT(NO_SUCH_NICK),      "There was no such nickname" },
+  { STAT(NO_SUCH_CHANNEL),   "There was no such channel" },
+  { STAT(NO_SUCH_SERVER),    "There was no such server" },
+  { STAT(INCOMPLETE_INFORMATION),  "Incomplete registration information" },
+  { STAT(NO_RECIPIENT),      "No recipient given" },
+  { STAT(UNKNOWN_COMMAND),   "Unknown command" },
+  { STAT(WILDCARDS),         "Wilcrads not allowed" },
+  { STAT(NO_CLIENT_ID),      "No Client ID given" },
+  { STAT(NO_CHANNEL_ID),     "No Channel ID given" },
+  { STAT(NO_SERVER_ID),      "No Server ID given" },
+  { STAT(BAD_CLIENT_ID),     "Bad Client ID" },
+  { STAT(BAD_CHANNEL_ID),    "Bad Channel ID" },
+  { STAT(NO_SUCH_CLIENT_ID), "There is no such client" },
+  { STAT(NO_SUCH_CHANNEL_ID),"There is no such channel" },
+  { STAT(NICKNAME_IN_USE),   "Nickname already exists" },
+  { STAT(NOT_ON_CHANNEL),    "You are not on that channel" },
+  { STAT(USER_NOT_ON_CHANNEL),"They are not on the channel" },
+  { STAT(USER_ON_CHANNEL),   "User already on the channel" },
+  { STAT(NOT_REGISTERED),    "You have not registered" },
+  { STAT(NOT_ENOUGH_PARAMS), "Not enough parameters" },
+  { STAT(TOO_MANY_PARAMS),   "Too many parameters" },
+  { STAT(PERM_DENIED),       "Permission denied" },
+  { STAT(BANNED_FROM_SERVER),"You are banned from this server" },
+  { STAT(BAD_PASSWORD),      "Cannot join channel. Incorrect password" },
+  { STAT(CHANNEL_IS_FULL),   "Cannot join channel. Channel is full" },
+  { STAT(NOT_INVITED),     "Cannot join channel. You have not been invited" },
+  { STAT(BANNED_FROM_CHANNEL), "Cannot join channel. You have been banned" },
+  { STAT(UNKNOWN_MODE),    "Unknown mode" },
+  { STAT(NOT_YOU),         "Cannot change mode for other users" },
+  { STAT(NO_CHANNEL_PRIV), "Permission denied. You are not channel operator" },
+  { STAT(NO_CHANNEL_FOPRIV),"Permission denied. You are not channel founder" },
+  { STAT(NO_SERVER_PRIV),  "Permission denied. You are not server operator" },
+  { STAT(NO_ROUTER_PRIV),  "Permission denied. You are not SILC operator" },
+  { STAT(BAD_NICKNAME),    "Bad nickname" },
+  { STAT(BAD_CHANNEL),     "Bad channel name" },
+  { STAT(AUTH_FAILED),     "Authentication failed" },
+  { STAT(UNKNOWN_ALGORITHM), "Unsupported algorithm" },
+  { STAT(NO_SUCH_SERVER_ID), "No such Server ID" },
+  { STAT(RESOURCE_LIMIT), "No more free resources" },
+  { STAT(NO_SUCH_SERVICE), "Service doesn't exist" },
+  { STAT(NOT_AUTHENTICATED), "You have not been authenticated" },
+  { STAT(BAD_SERVER_ID), "Server ID is not valid" },
+  { STAT(KEY_EXCHANGE_FAILED), "Key exchange failed" },
+  { STAT(BAD_VERSION), "Bad version" },
+
+  { 0, NULL }
+};
+
+/* Returns status message string */
+
+char *silc_get_status_message(unsigned char status)
+{
+  int i;
+
+  for (i = 0; silc_status_messages[i].message; i++) {
+    if (silc_status_messages[i].status == status)
+      break;
+  }
+
+  if (silc_status_messages[i].message == NULL)
+    return "";
+
+  return silc_status_messages[i].message;
+}
